#include <stdint.h>
#include "config.h"
#include "def.h"
#include <ffft.h>

#define ROLL       0
#define PITCH      1
#define YAW        2

static int16_t capture[FFT_N];			/* Wave captureing buffer */
static complex_t bfly_buff[FFT_N];		/* FFT buffer */

typedef struct serial_buff_t{
  uint8_t  SOT[3];
  uint8_t  acc_type;
  uint16_t acc_1G;
  uint16_t sample_size;
  uint16_t spekt_size;
  uint16_t period;
  uint16_t i2c_errors_count;
  uint16_t enc_period;
  int16_t min_sample;
  int16_t max_sample;
  uint16_t spektrum[FFT_N/2];                  /* Spectrum output buffer */
};

static union {serial_buff_t val; uint8_t raw[sizeof(serial_buff_t)]; } serial_buff;

static uint16_t i2c_errors_count = 0;
static uint16_t calibratingG = 0;
static int16_t  gyroADC[3],accADC[3];
static int16_t  gyroZero[3] = {0,0,0};
static uint8_t  nunchuk = 0;
static uint16_t calibratingA = 0;  // the calibration is done is the main loop. Calibrating decreases at each cycle down to 0, then we enter in a normal mode.
static int16_t  accZero[3]  = {0,0,0};
static uint16_t acc_1G;             // this is the 1G measured acceleration

void blinkLED(uint8_t num, uint8_t wait,uint8_t repeat) {
  uint8_t i,r;
  for (r=0;r<repeat;r++) {
    for(i=0;i<num;i++) {
      LEDPIN_TOGGLE; //switch LEDPIN state
      BUZZERPIN_ON;
      delay(wait);
      BUZZERPIN_OFF;
    }
    delay(60);
  }
}

void writeParams() {
  blinkLED(10, 15, 5);
}  

void rxInt() {
  uint16_t now;
  static uint16_t last = 0;
  static uint8_t chan = 0;
  now = micros();
  serial_buff.val.enc_period = now - last;
  last = now;
}


void setup() {
  Serial.begin(115200);
  LEDPIN_PINMODE
  delay(200);
  POWERPIN_ON;
  delay(100);
  i2c_init();
  delay(100);
  ACC_init();
  PPM_PIN_INTERRUPT
  serial_buff.val.SOT[0] = '<';  
  serial_buff.val.SOT[1] = 'S';  
  serial_buff.val.SOT[2] = '>';  
  serial_buff.val.acc_type = 1;  
  serial_buff.val.spekt_size = FFT_N/2;  
  serial_buff.val.sample_size = FFT_N;
  serial_buff.val.acc_1G = acc_1G;
  calibratingA = 400;
  acc_1G = 0;
}

void MinMax() {
  serial_buff.val.min_sample = capture[0];
  serial_buff.val.max_sample = capture[0];
  for(uint16_t i=1; i<FFT_N; i++) {
    if (serial_buff.val.min_sample > capture[i])   
      serial_buff.val.min_sample = capture[i];
    if (serial_buff.val.max_sample < capture[i])   
      serial_buff.val.max_sample = capture[i];
  }  
}

void loop() {
  uint32_t p_start;
  uint32_t p_end;
  if (!calibratingA) {  
    i2c_errors_count = 0;
    p_start = micros();
    for(uint16_t i=0; i<FFT_N; i++) {
      ACC_getADC();
      capture[i] = accADC[YAW]; 
    }
    p_end = micros();
    MinMax();
    fft_input(capture, bfly_buff);
    fft_execute(bfly_buff);
    UartFlushBuff();
    serial_buff.val.period = (p_end - p_start)/4;
    serial_buff.val.i2c_errors_count = i2c_errors_count;
    fft_output(bfly_buff, serial_buff.val.spektrum);
    UartSendData();
    LEDPIN_TOGGLE;
  } else {
    ACC_getADC();
  }  
}




