import processing.core.*; 
import processing.xml.*; 

import processing.serial.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class VibAnalyser extends PApplet {

//
int frame_size = 115;
/*
typedef struct serial_buff_t{
  uint8_t  SOT[3];
  uint8_t  acc_type;
  uint16_t acc_1G;
  uint16_t sample_size;
  uint16_t spekt_size;
  uint16_t period;
  uint16_t i2c_errors_count;
  uint16_t enc_period;
  int16_t min_sample;
  int16_t max_sample;
  uint16_t spektrum[FFT_N/2];            
};
*/
int acc_type  = 0;
int acc_1G = 0;
int sample_size = 0;
int spekt_size = 0;
int period = 0;
int i2c_errors_count = 0;
int enc_period = 0;
int min_sample = 0;
int max_sample = 0;
// Feel Free to edit these variables ///////////////////////////
String  xLabel = "Frequency, Hz";
String  yLabel = "Acceleration, g";
String  Heading = "Vibration Analyser";
String  URL = "http://www.multiwii.com/forum/";
int NumOfVertDivisions=5;      // dark gray
int NumOfVertSubDivisions=10;  // light gray
int NumOfBars=64;    // you can choose the number of bars, but it can cause issues  
                    // since you should change what the arduino sends
float BarLPF = 0.3f;  
float AmpLPF = 0.25f;  
float BaseFreq = 3200.0f;
float RpmFreq = 10.0f;
float CogFreq = 10.0f;
float PekAmp = 0.0f;

// if these are changed, backgroung image has problems 
// a plain background solves the problem
int ScreenWidth = 800, ScreenHeight=600;
/////////////////////////////////////////////////////////

//  Serial port stuff ///////////////////////

Serial g_serial;        
boolean firstContact = false; 
int[] serialInArray = new int[NumOfBars];
int serialCount = 0;
///////////////////////////////////////////////

int LeftMargin=100;
int RightMArgin=80;
int TextGap=60;
int GraphYposition=80; 
float BarPercent = 0.4f;

int value;

PFont font;
PImage bg;

int temp;
float yRatio = 0.58f;
int BarGap, BarWidth, DivisounsWidth;
float[] bars = new float[NumOfBars];

public void setup(){
  bg = loadImage("BG.jpg"); 
  /// NB SETTINGS ////////////////////////////////////////////////////////
  g_serial = new Serial(this, Serial.list()[0], 115200); 
  g_serial.buffer(20 + NumOfBars*2);
  ////////////////////////////////////////////////////////////////////////
  DivisounsWidth = (ScreenWidth-LeftMargin-RightMArgin)/(NumOfBars); 
  BarWidth = PApplet.parseInt(BarPercent*PApplet.parseFloat(DivisounsWidth));
  BarGap = DivisounsWidth - BarWidth;
  size(ScreenWidth,ScreenHeight);
  font = createFont("Tahoma", 12);
  textAlign(CENTER);
  textFont(font);
}

public void draw(){
  background(250);      
  Axis();
  Labels();
  PrintBars();
  Headings();           
}

public int reads16() {return (PApplet.parseByte(g_serial.read())&0xff) + (PApplet.parseByte(g_serial.read())<<8);}
public int readu16() {return (g_serial.read()&0xff) + (g_serial.read()<<8);}
public int readu8()  {return g_serial.read()&0xff;}


public void processSerialData() {
  if ((g_serial.read() == 'S') && (g_serial.read() == '>')) {
    acc_type = readu8();
    acc_1G = readu16();
    sample_size = readu16();
    spekt_size = readu16();
    period = readu16();
    i2c_errors_count = readu16();
    enc_period = readu16();
    min_sample = reads16();
    max_sample = reads16();
    for (int i=0; i<NumOfBars; i++)
      serialInArray[i] = readu16();
    g_serial.read();
    for (int x=0;x<NumOfBars;x++)
      bars[x] = bars[x]*(1 - BarLPF) + BarLPF * yRatio*(ScreenHeight)*(serialInArray[x]*1.0f/acc_1G);
    BaseFreq = round(1000000.0f/period/4.0f)*sample_size;  
    if (enc_period == 0) 
      enc_period = 65535;
    RpmFreq = round(100000000.0f/enc_period/7.0f)/100.0f;
    CogFreq = RpmFreq*12.0f;
    PekAmp  = (max_sample-min_sample)*1.0f/acc_1G;
    //PekAmp * (1 - AmpLPF) + AmpLPF*round((max_sample-min_sample)*10.0/acc_1G)/10.0;
    //println(1000.0/period*sample_size);  
    //println(i2c_errors_count);
    //println(1000000.0/enc_period*60.0/7.0);
    //println(min_sample);
    //println(CogFreq);
  } else {g_serial.readStringUntil('<'); println("NO_SYNC");} ;
}

// Send Recieve data //
public void serialEvent(Serial myPort) {
  processSerialData();
}

/////// Display any variables for testing here//////////////
public void Headings(){
  textFont(font,12);
  textAlign(LEFT);
  fill(0);
  text("RPM",50,TextGap );   
  text(round(RpmFreq*60),100,TextGap );    
  fill(0, 255, 0);
  text("Frpm",50,TextGap + 20 );   
  text(round(RpmFreq),100,TextGap + 20);    
  fill(255, 0, 0);
  text("Fcog",50,TextGap + 40 );   
  text(round(CogFreq),100,TextGap + 40);    

  fill(0);
  text("Fadc",150,TextGap );   
  text(round(BaseFreq),200,TextGap );    
  text("Amp.",150,TextGap + 20 );   
  text(nf(PekAmp, 0, 2),200,TextGap + 20);    
  text("i2c err.",150,TextGap + 40 );   
  text(i2c_errors_count,200,TextGap + 40);    
}

public void PrintBars(){ 
  textFont(font,8);
  int c=0;
  for (int i=0;i<NumOfBars;i++){
    fill((0xe4+c),(255-bars[i]+c),(0x1a+c));
    stroke(90);
    rect(i*DivisounsWidth+LeftMargin,  ScreenHeight-GraphYposition,   BarWidth,   -bars[i]);
    fill(0x2e,0x2a,0x2a);
  }
  strokeWeight(3);
  stroke(0, 255, 0, 150);
  float pos = (NumOfBars-1)*DivisounsWidth/BaseFreq*2.0f*RpmFreq+LeftMargin-DivisounsWidth+BarWidth;
  line(pos, ScreenHeight-GraphYposition, pos, ScreenHeight-GraphYposition - yRatio*(ScreenHeight));
  stroke(255, 0, 0, 150);
  pos = (NumOfBars-1)*DivisounsWidth/BaseFreq*2.0f*CogFreq+LeftMargin-DivisounsWidth+BarWidth;
  line(pos, ScreenHeight-GraphYposition, pos, ScreenHeight-GraphYposition - yRatio*(ScreenHeight));
  strokeWeight(1);
}

public void Axis(){
  strokeWeight(1);
  stroke(220);
  for(float x=0;x<=NumOfVertSubDivisions;x++){
    int bars=(ScreenHeight-GraphYposition)-PApplet.parseInt(yRatio*(ScreenHeight)*(x/NumOfVertSubDivisions));
    line(LeftMargin-15,bars,ScreenWidth-RightMArgin-DivisounsWidth+50,bars);
  }
  strokeWeight(1);
  stroke(180);
  for(float x=0;x<=NumOfVertDivisions;x++){
    int bars=(ScreenHeight-GraphYposition)-PApplet.parseInt(yRatio*(ScreenHeight)*(x/NumOfVertDivisions));
    line(LeftMargin-15,bars,ScreenWidth-RightMArgin-DivisounsWidth+50,bars);
  }
  strokeWeight(2);
  stroke(90);
  line(LeftMargin-15, ScreenHeight-GraphYposition+2, ScreenWidth-RightMArgin-DivisounsWidth+50, ScreenHeight-GraphYposition+2);
  line(LeftMargin-15,ScreenHeight-GraphYposition+2,LeftMargin-15,GraphYposition+80);
  strokeWeight(1);
}

public void Labels(){
  textAlign(CENTER);
  textFont(font,18);
  fill(50);
  rotate(radians(-90));
  text(yLabel,-ScreenHeight/2,LeftMargin-45);
  textFont(font,10);
  for(float x=0;x<=NumOfVertDivisions;x++){
    int bars=(ScreenHeight-GraphYposition)-PApplet.parseInt(yRatio*(ScreenHeight)*(x/NumOfVertDivisions));
    text(x / NumOfVertDivisions,-bars,LeftMargin-20);
  }
  textFont(font,18);
  rotate(radians(90));  
  text(xLabel,LeftMargin+(ScreenWidth-LeftMargin-RightMArgin-50)/2,ScreenHeight-GraphYposition+40);
  textFont(font,24);
  fill(50);
  text(Heading,LeftMargin+(ScreenWidth-LeftMargin-RightMArgin)/2,30);
  textAlign(RIGHT);
  textFont(font);
  fill(150);
  text(URL,ScreenWidth-RightMArgin,ScreenHeight-15);
  textAlign(RIGHT);
  textFont(font, 8);
  rotate(radians(-90));
  for (int i=0;i<NumOfBars;i++){
    text(round(BaseFreq*(i+1)/NumOfBars/2), GraphYposition-ScreenHeight-10, i*DivisounsWidth+LeftMargin + 4);
  } 
  rotate(radians(90));
  textFont(font);
}


  static public void main(String args[]) {
    PApplet.main(new String[] { "--bgcolor=#F0F0F0", "VibAnalyser" });
  }
}
